/********************************************************/
/* Common Device Interface definition file */
/* ------------ cdi.h --------------- */
/********************************************************/

#ifndef CDI_H_
#define CDI_H_

#include "cdi.def"

#ifndef CDI_EXPORT
# define CDI_EXPORT
#endif
#include "tine.h"
#include "cdi_prot.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

extern int cdiHeapCorrupted;
#ifdef WIN32
# define cdiAssertHeapOkay(t,d,p) {\
  if (!cdiHeapCorrupted && _heapchk() != _HEAPOK) {\
    cdilog("%s %s access: %x : memory overwrite has occurred",t,d,p);\
    cdiHeapCorrupted = TRUE;\
  }\
}
#else
# define cdiAssertHeapOkay(t,d,p) ;
#endif
#ifdef UNIX
typedef void* HANDLE;
#endif

extern time_t gCdiStartupTime;

#define CDI_BUSNAME_SIZE 16
#define CDI_DEVICE_NAME_SIZE 32
#define CDI_DEVICE_LONGNAME_SIZE 64
#define CDI_TEMPLATE_NAME_SIZE 16
#define CDI_TEMPLATE_FIELDNAME_SIZE 16
#define CDI_DEVICE_DESC_SIZE 64
#define CDI_PARAM_NAME_SIZE 64
#define CDI_PATTERN_FILTER_SIZE 32
#define CDI_ADDR_SIZE 16
#define CDI_UNITS_SIZE 16

#define CDI_PAYOFF_HASHSIZE 25
#define CDI_TINY_HASHSIZE 31
#define CDI_SMALL_HASHSIZE 251
#define CDI_MEDIUM_HASHSIZE 1021
#define CDI_LARGE_HASHSIZE 4093

extern cdi_mutex_t logMutex;

typedef struct cdiHashXRef
{
  int cdiLine;
  int cdiIndex;
  int devNumber;
  struct cdiHashXRef *nxt;
} cdiHashXRefStruct;

#define DEF_INPUT_BYTES 32  /* default input bytes for input */
#define MAX_REDIRECTED_DEVICES 256
typedef struct cdiRedirXRef
{
  char devName[CDI_DEVICE_NAME_SIZE];
  char devRedirectedServerName[64];
  int devNumber;
  struct cdiRedirXRef *nxt;
} cdiRedirXRefStruct;

typedef struct cdiBitfieldXRefStruct
{
  DBITFIELD *bf;
  char device[CDI_DEVICE_NAME_SIZE];
  struct cdiBitfieldXRefStruct *nxt;
} cdiBitfieldXRef;
extern cdiBitfieldXRef *gCdiBitfieldXRefLst;
#define BITFIELD_LCL "thisfec"

#define BACK_THREAD_TIME 10000 /* 10 ms for usleep */
#define MULTI_N_DEVICE 5 /* allocate multiple RequestBlocks */
#define MULTI_MAX_DEVICE 20 /* allocate multiple RequestBlocks */
#define NUMBER_LINES 20 /* total number of bus lines */

#define NUMBER_EXEC_PARAM 16 /* number parameters for exec and poll Request call */

#define DATA_BACK 1
#define DATA_PASS_FILTER 2

#define THREAD_INDEX_MASTER		1
#define THREAD_INDEX_CALLBACK	2

#define FREE_ONCLOSE     0x4000
#define RETURN_NAMES     0x2000
#define RETURN_STATUS    0x1000
#define RETURN_CALIBRATED 0x800
#define RETURN_REFORMATED 0x400
#define INPUT_CALIBRATED  0x200
#define INPUT_REFORMATED  0x100

#define CDICHAR_ADD '+'
#define CDICHAR_SUB '-'
#define CDICHAR_MUL '*'
#define CDICHAR_DIV '/'
#define CDICHAR_PWR '^'
#define CDICHAR_LOG 'L'
#define CDICHAR_MOD '%'
#define CDICHAR_RGH '>'
#define CDICHAR_LFT '<'
#define CDICHAR_UNS 'U'
#define CDICHAR_SGN 'S'
#define CDICHAR_MSG 'M'
#define CDICHAR_EXP 'E'
#define CDICHAR_FCN '|'
#define CDICHAR_FCNA 'F'
#define CDICHAR_XOR 'X'
#define CDICHAR_NOT '~'
#define CDICHAR_AND '&'
#define CDICHAR_FIX '='
#define CDICHAR_OR  'O'

enum cdiRuleOperation
{
  cdiRuleAdd,
  cdiRuleSub,
  cdiRuleMult,
  cdiRuleDiv,
  cdiRulePwr,
  cdiRuleLog,
  cdiRuleMod,
  cdiRuleShiftRight,
  cdiRuleShiftLeft,
  cdiRuleUnsigned,
  cdiRuleSigned,
  cdiRuleMessage,
  cdiRuleExp,
  cdiRuleFcn,
  cdiRuleXor,
  cdiRuleNot,
  cdiRuleAnd,
  cdiRuleFix,
  cdiRuleOr
};

typedef struct cdiRuleStruct
{
  int operation;
  int isSigned;
  int msgOnValue;
  int reserved;
  double operand;
  double (*fcn)(double);
  char msgOn[32];
  char msgOff[32];
  struct cdiRuleStruct *nxt;
} CdiRule;

typedef struct cdiClbrFcnStruct
{
  char fcnName[64];
  double (*fcn)(double);
  struct cdiClbrFcnStruct *next;
} CdiClbrFcn;

extern CdiClbrFcn *cdiClbrFcnLst;
int registerClbrConversionFunction(char *fcnName,double (*fcn)(double));

typedef struct CdiTemplateListStruct
{
  char field[CDI_TEMPLATE_FIELDNAME_SIZE]; /* field name to be applied */
  char lngfield[64]; /* long field name to be applied */
  char description[96]; /* field description */
  int  addrParameters[NUMBER_EXEC_PARAM]; /* relative offsets */
  int  naddrParameters;
  BYTE *dataMask;  /* default data mask */
  int  dataMaskRange[2];
  BYTE *dataPattern;/* default data tolerance/pattern */
  int  dataPatternRange[2];
  int  dataPatternTrigger;
  BYTE *dataInput;   /* default data input for WRITE commands */
  int dataInputSize; /* maximum allowed data input size in bytes (if > 0) */
  int dataOutputSize; /* maximum allowed data output size in bytes (if > 0) */
  int dataOutput2ndSize; /* maximum allowed data output size in bytes (if > 0) */
  int dataFormat;    /* default data type */
  int allowedAccess; /* allowed bus access flag */
  int interval; /* default hardware interval */
  CdiRule *clbrRules;/* associated calibration rules for device readout */
  CdiRule *rvrsRules;/* associated reverse calibration rules for device sendto */
  DBITFIELD *bitfield; /* bitfield reference if not NULL */
  struct CdiTemplateListStruct *next;
} CdiTemplateList;

typedef struct CdiTemplateInstanceNameStruct
{
  char name[32];
  struct CdiTemplateInstanceNameStruct *nxt;
} CdiTemplateInstanceName;
typedef struct 
{
  char name[CDI_TEMPLATE_NAME_SIZE];
  int length;
  int numInstances;
  CdiTemplateInstanceName *instanceNames;
  CdiTemplateList *cdiTemplateList;
} CdiTemplate;

#define MAX_NUM_TEMPLATES 64
extern CdiTemplate *CdiTemplateTbl[MAX_NUM_TEMPLATES];
extern int CdiTemplateTblSize;
CdiTemplate *getCdiTemplate(char *templateName);
CdiTemplateList *findFieldInTemplate(CdiTemplate *tmpt,char *field);
int addFieldToTemplate(
  CdiTemplate *tmpt, char *field, char *lngfield, char *fieldDesc,
  int *addrParameters,int naddrParameters,
  BYTE *mask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger,
  BYTE *dataIn,int maxBytesIn,int maxBytesOut, int maxBytes2ndOut, int dataFmt,
  int access,
  int interval,
  CdiRule *rules_rcv, CdiRule *rules_snd,
  DBITFIELD *bf
);

/**
 * \brief CDI Device Information Block structure
 * 
 * This is the principal device structure passed to and from CDI and a bus plug
 * 
 */
typedef struct
{
  char devName[CDI_DEVICE_NAME_SIZE]; /**< registered (hashed) device name */
  int cdiIndex;      /**< start from 0, for pDev[] */
  int devNumber;     /**< start from 1, unique number */
  void *pcdiLine;    /**<  pointer to CdiLineInfoBlk (before: int cdiLine; index for CdiLineInfoBlk) */
  int cdiAddr[CDI_ADDR_SIZE]; /**< module address specification */
  int numberDeviceError; /**< total number of bus errors */
  int *devParameters;/**< default bus parameters for this device */
  BYTE *dataMask;    /**< default data mask */
  int dataMaskStartIndex; /**< list start index */
  int dataMaskStopIndex; /**< list stop index */
  BYTE *dataPattern; /**< default data tolerance/pattern */
  int dataPatternStartIndex; /**< list start index */
  int dataPatternStopIndex; /**< list stop index */
  int dataPatternTrigger; /**< seek data pattern if TRUE */
  BYTE *dataInput;   /**< default data input for WRITE commands */
  int dataInputSize; /**< maximum allowed data input size in bytes (if > 0) */
  int dataOuputSize; /**< maximum allowed data output size in bytes (if > 0) */
  int dataOuput2ndSize; /**< maximum allowed Second data output size in bytes (if > 0) for RR & WRR mode */
  int dataFormat;    /**< default data type */
  int reqFormat;     /**< requested data type */
  int allowedAccess; /**< allowed bus access flag */
  int interval;      /**< default hardware interval */
  int iReserved;     /**< reserved for future use */
  void *ptrReserved; /**< reserved for future use */
  CdiRule *clbrRules;/**< associated calibration rules for device readout */
  CdiRule *rvrsRules;/**< associated reverse calibration rules for device sendto */
  char lngName[CDI_DEVICE_LONGNAME_SIZE];/**< associated long device name */
  char description[CDI_DEVICE_DESC_SIZE];/**< associated device description */
  float maximum;     /**< display maximum */
  float minimum;     /**< display minimum */
  char precision[CDI_UNITS_SIZE];/**< display format a la width.decimal */ 
  char units[CDI_UNITS_SIZE];    /**< display units (or bitfield tag) */
} CdiDeviceInfoBlk;
#define DEVICE_INFOR_SIZE sizeof(CdiDeviceInfoBlk)

typedef struct
{
  char paramName[CDI_PARAM_NAME_SIZE];
  int devParameters[NUMBER_EXEC_PARAM];
} PARAM_STRUCT;

/**
 * \brief CDI Request Information Block structure
 * 
 * This is the principal structure passed to and from CDI and a bus plug
 * 
 */
typedef struct
{
  int active;	            /**< number of request process steps */
  int callBackIndex;      /**< can be user assigned (< 0x1000) or requestNumber + 0x1000 */
  CdiDeviceInfoBlk *pDev;	/**< referenced device */
  void *pInData;		      /**< pointer to data coming from the device */
  unsigned int inLength;  /**< size in bytes of data coming from the device */
  void *pOutData;		      /**< pointer to data sent to the device */
  unsigned int outLength; /**< size in bytes of data sent to the device */
  short inType;           /**< data type (from the hardware) one of char, short, int, float or double	*/
  short outType;	        /**< data type (to the hardware) one of char, short, int, float or double	*/
  short inFormatCode;     /**< original caller's format code */
  short outFormatCode;    /**< original caller's format code */
  int hasPatternFilter;   /**< 0 = FALSE (no pattern); else TRUE */
  BYTE patternFilter[CDI_PATTERN_FILTER_SIZE]; /**< compare pattern or Filter, if string max 32 bytes, otherwise check 8 bytes(double) only */
  int hasDataMask;        /**< 0 = FALSE (no mask); else TRUE */
  BYTE mask[8];           /**< data mask to apply */
  int maskValue;          /**< ?? */
  int devParameters[NUMBER_EXEC_PARAM]; /**< address parameters registered with this device */
  int timeout;            /**< in milliseconds: < 0 return only on time out or error; 0: wait forever */
  int registerTime;       /**< time when contract was registered */
  int requestTime;        /**< time whem contract was last sent */
  int pollTime;           /**< polling interval ? */
  int requestCounter;     /**< number of requests */
  short accessFlag;       /**< read 1, write 2, read then write 3, write then read 4 */
	short	conditionFlag;    /**< for reading data (readConditionNumeraters), bit: 0x8 for mask ON flag */
  int needsCalibration;   /**< if TRUE -> needs to calibrate results */
  int requestNumber;      /**<  reqIndex + (cdiLine << 16) */
  int reqIndex;		        /**< own index number: 0, 1... */
  void *groupReq;	        /**< grouped request, group true if != NULL */
  short NgroupsReqs;		  /**< number of connected requests for mainRequest, but index number for each requested device start from 1  */
	short NgroupRets;		    /**< number of data returnd or timeout... */
  short NgrpErrors;       /**< number of errors of this group  */
  short groupSyncStart;   /**< synchronous flag for group start */
  short dataReturnCondition; /**< databack: 1; conditionOK: 2 */
	short asynCallFlag;     /**< flag to indicate whether the call is asynchronous */
  int deviceError;        /**< 0 if no bus error on the device */
  BYTE  cpOutData[DEF_INPUT_BYTES];	/**< default data to send to the bus */
  short bufLastData[4];	  /**< comparison buffer	*/
  int errorCode, *perror; /**< specific error code; table of errors */
  int responseTime;       /**< last response time in ms */
  short pollFlag;         /**< infinite loop for async call */
	short lockRequestFlag;	/**< lock this request until cancel calls	*/
	short userCallDown;     /**< callback or syncSem done flag */
	short cancelFlag;		    /**< cancel callback */
  sem_t sync_SemId;       /**< bus request semaphores */
  sem_t exec_SemId;       /**< bus request semaphores */
  int bufIndex;           /**< a buffer index used by the bus plug  */
  void (*callback)(int, int); /**< callback function: receivces callBackIndex, errorCode */
  int freeInDataMemoryOnCancel; /**< memory managment flag */
  int freeOutDataMemoryOnCancel;/**< memory managment flag */
  int inside;             /**< synchronization flag */
  void  *pBlockBuffer;    /**< block transfer buffer pointer */
  int devReadyTime;       /**< polling device ready  */
  int nerrors;            /**< number of consecutive bus errors */
  void *ptrReserved;      /**< reserved for future use */
} CdiRequestInfoBlk;
#define REQUEST_INFOR_SIZE sizeof(CdiRequestInfoBlk)

typedef struct
{
  int busIndex; /* index into CdiBusInfoBlk table */
  int cdiLine; /* line number for infor Struct, start from 0,1,2 */
  int busLine; /* bus line number, e.g sedac line 1,2,3,4... */
  int fileHandle; /* opened file handle */
  int numberRequest;
  int numberActiveRequest;
  int numberDevices;
  int backCounter, activeCounter;
  int readCounter, callCounter;
  struct timeval tv;
  int backTime; /* background task timer in msec */
  int nextWakeupTime;
  struct timespec backReq;
  CdiDeviceInfoBlk *pDevList;
  CdiRequestInfoBlk *pReqList;
  cdiHashXRefStruct **devHashList;
  int devHashSize;
  int threadActive;
  int threadBusy;
  int iReserved;     /* reserved for future use */
  void *ptrReserved; /* reserved for future use */
  sem_t back_SemId, call_SemId;
  cdi_mutex_t lineMutex; /* for user- and background-task */
  cdi_thread_t tIdMaster; /* background task */
  cdi_thread_t tIdCallback; /* callback task */
} CdiLineInfoBlk;
#define LINE_INFOR_SIZE sizeof(CdiLineInfoBlk)

typedef struct
{
  char busName[CDI_BUSNAME_SIZE];
  int busType;
  int iReserved;     /* reserved for future use */
  void *ptrReserved; /* reserved for future use */
  int (*filterBusFcn)(int *parameter, int accessFlag);/* associated bus parameter filter */
  int (*cfgBusFcn)(void);/* associated bus configurer */
  int (*initBusFcn)(int busline,int cdiline,int ndevices,char *paramList);/* associated bus initializer */
  void (*execBusFcn)(CdiRequestInfoBlk *pReq); /* associated bus executer */
  int (*freeBusFcn)(int busline);             /* associated bus disconnecter */
  int (*scanBusFcn)(char *target,char *result,int resultBufferSize); /* associated bus scanning function */
  int (*blockFcn)(int *requestIndex, int numberRequest, void **blockBuffer, int openFlag); /* associated bus block transfer function */
  int (*timeoutFcn)(CdiRequestInfoBlk *pReq, int set, int taskIndex);/* associated bus timeout function */
  int (*deviceReadyFcn)(CdiRequestInfoBlk *pReq, int taskIndex);/* associated bus device Ready function */
  int (*loadRequestFcn)(CdiRequestInfoBlk *pReq, int taskIndex);/* associated load request function */
  int (*cancelRequestFcn)(CdiRequestInfoBlk *pReq, int taskIndex);/* associated cancel request function */
} CdiBusInfoBlk;

enum cdiAddrNumeraters /* parameters for cdiAddr */
{
  caAddrIndex = 0,
  subAddrIndex,
};

enum taskIndexNumeraters /* parameters for call cdiExecRequest() */
{
  readTaskIndex,
  backTaskIndex,
  execIndex,
  cancelIndex,
};

enum reqStatus /* parameters for requestActive flag */
{
  idleState = 1,		/* after registration	 */
  transRequest,
  transPoll,
  recvRequest,
  transRequest2nd,    /* start from backgroud task immediately */
  transPoll2nd,
  recvRequest2nd,
  finishRequest,
  removeRequest,
};

enum cdiDataType	/* cdi support primitive data type */
{
  cdiChar,
  cdiByte,
  cdiShort,
  cdiInteger,
  cdiFloat,
  cdiDouble,
  cdiChar32,
};

extern PARAM_STRUCT *paramList;
extern int paramListSize;
extern int numBusList;
extern CdiBusInfoBlk busList[NUMBER_LINES];
extern CdiLineInfoBlk *busHandle[NUMBER_LINES];
extern int CdiNumIndiciesPerLine[NUMBER_LINES];
extern int numConnectedHandles;
extern int cdiInitFlag;
extern CDI_EXPORT int cdiDebugLevel;
extern CDI_EXPORT int cdiBusPlugDebugLevel;

CDI_EXPORT THREAD_TYPE ThreadMaster(THREAD_PARAM param);
CDI_EXPORT THREAD_TYPE ThreadCallback(THREAD_PARAM param);

extern cdiRedirXRefStruct **cdiRedirXref;
extern int nCdiRedirectedDevices;
extern int maxCdiRedirectedDevices;
extern int gCdiRedirectedDevicesHashSize;
int cdiDumpLineInfo(int i1,int i2,int i3,int i4);
int cdiGetNextAvailableBusType(void);
int getNextFreeIndex(int cdiLine);
int cdiGetBusType(char *busName);
int cdiRemapDataSets(DTYPE *dout,BYTE *dataout,int *errLst,NAME32 *nameout);
int cdiParseDeviceString(char *devString,NAME32 *devGroup,int *groupSize);
CDI_EXPORT int findDevice(char *devName, int *pCdiLine, int *pDevIndex);
CDI_EXPORT void checkReturnCondition(CdiRequestInfoBlk *pReq, char *lastData, char *str, int datalen, int stringType);
CDI_EXPORT void cdiExecRequest(CdiRequestInfoBlk *pReq, int taskIndex);
/* supported bus type: */
/* BusType: "SEDAC", "RS232", "CAN", "GPIB" */
CDI_EXPORT int cdi_Open(char *busName, int busLine, char *parameter, int numberDevices, int numberRequestBuffers);
CDI_EXPORT int cdi_Close(int cdiLine);
CDI_EXPORT int cdi_RegisterDevice(
  int cdiLine, char *devName, char *lngName, char *devDesc,
  int cdiIndex, int devNumber, int *cdiAddress, int cdiAddressLength
  );
CDI_EXPORT int cdi_RegisterParameters(char *paramName, int *parameters);
CDI_EXPORT int *cdi_GetParameters(char *paramName);
/* get routines mainly for statistics */ 
CDI_EXPORT int cdi_GetCdiLine(char *busName, int busLine); /* get cdiLine index from harware bus name("SEDAC") and line number(sedac line number)  */ 
CDI_EXPORT int cdi_GetNumberRegisteredBus(void);
CDI_EXPORT int cdi_GetNumberRequestsMax(int cdiLine);  /* cdicdiLine */
CDI_EXPORT int cdi_GetNumberRequests(int cdiLine);  /* cdicdiLine */
CDI_EXPORT int cdi_GetNumberActiveRequests(int cdiLine);  /* cdicdiLine */
CDI_EXPORT int cdi_GetActiveRequestInfor(int cdiLine, int *pBuffer, int size, int offset);
CDI_EXPORT int checkDeadGroupCalls(CdiRequestInfoBlk *pReq);
CDI_EXPORT int cdi_GetNumberOpenedLines(void);  /* totol opened lines */
CDI_EXPORT int cdi_GetNumberConnectedHandles(void);
CDI_EXPORT int cdi_GetNumberOpenedLinesByBus(char *busName);  /* for this type of bus */
CDI_EXPORT int cdi_GetNumberDevicesByLine(int cdiLine);  /* cdicdiLine */
CDI_EXPORT int cdi_GetRequestErrors(int requestIndex, int *errorArray);
CDI_EXPORT int cdi_GetRequestDeviceNames(int requestIndex, NAME32 *namebuffer);
CDI_EXPORT int cdi_GetRequestStat(int requestIndex, int *activeStatArray);
CDI_EXPORT int cdi_GetRequestNumberDevices(int requestIndex, int *numberOfDevices);
CDI_EXPORT int cdi_Request(
  char *devName, void *readData, short readType, unsigned int readLength, void *writeData, short writeType, 
  unsigned int writeLength, void *dataPatternWindow, void *dataMask, int *parameter, 
  int accessFlag, int returnCondition, int timeout, void (*callBack)(int, int), 
  int pollFlag, int *perror, int callbackID, int lockRequest
  );
CDI_EXPORT int cdi_loadGroupRequests(int requestIndexMain, int requestIndexFollower);
CDI_EXPORT int cdi_isValidRequestIndex(int requestIndex);
CDI_EXPORT int cdi_CancelRequest(int requestIndex);
CDI_EXPORT int cdi_RequestTerminated(int requestIndex);
CDI_EXPORT int cdi_StartRequest(int requestIndex);
CDI_EXPORT int cdi_StopRequest(int requestIndex);
CDI_EXPORT int cdi_assignDefaultDeviceParameters(char *devName,int *devParameters);
CDI_EXPORT int cdi_assignDefaultDataMaskAndPattern(char *devName,BYTE *dataMask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger);
CDI_EXPORT int cdi_assignAllowedAccess(char *devName,int access);
CDI_EXPORT int cdi_eraseDefaultDataMaskAndPattern(char *devName,int erasedataMask, int erasepattern);

CDI_EXPORT int cdi_getDeviceNames(NAME32 *namebuffer, int namesize); /* for all cdi lines  */
CDI_EXPORT int cdi_getDeviceNamesByLine(NAME32 *namebuffer, int namesize, int cdiLine);  /* cdiLine is the cdiLineIndex  */
CDI_EXPORT char *cdi_getBusNameForCdiLine(int cdiLine);
CDI_EXPORT char *cdi_getBusNameForDevice(char *devName);
CDI_EXPORT char *cdi_getFieldBusNameForDevice(char *devName);
CDI_EXPORT int cdi_getBusNames(NAME16 *namebuffer,int namesize);
CDI_EXPORT int cdi_getDeviceNumberOfErrors(char *devName);
CDI_EXPORT DADDRESS *cdi_getBusAddressForDevice(char *devName);
CDI_EXPORT char *cdiGetRedirectedDeviceServer(char *devName);
CDI_EXPORT int cdi_getRedirectedDeviceNames(NAME32 *namebuffer,int namesize);
CDI_EXPORT int cdi_RegisterRedirectedDevice(
  char *devName, int devNumber, char *redirectedServerName
  );
CDI_EXPORT int cdi_RegisterBus(
  char *busName,
  int busType,
  int (*filterBusFcn)(int *parameter, int accessFlag),
  int (*cfgBusFcn)(void),
  int (*initBusFcn)(int busline,int cdiline,int ndevices,char *paramList),
  void (*execBusFcn)(CdiRequestInfoBlk *pReq),
  int (*freeBusFcn)(int busline)
  );
CDI_EXPORT int cdiRegisterBusFilter(char *busName,int (*fcn)(int *,int));
CDI_EXPORT int cdiRegisterBusConfiguration(char *busName,int (*fcn)(void));

/**
 * \brief registers a bus initialization function to be used by CDI when initializing the bus.
 * 
 * After the CDI database has been read and processed, CDI will call any bus initialization routines
 * which need to be called, so as to set up the hardware bus for hardware io.
 * 
 * \param busName is the bus plug name to be used by database entries refering to this bus plug.
 * \param fcn is the address of the bus initialization function to be called when CDI initializes the
 * bus. The function must have the prototype 
 * \n   int fcn(int busLine, int cdiLine, int numberDevices, char *parameterList)
 * \n and return '0' if the initialization was successful.  This initialization dispatch function
 * will receive (according to the registered database entries) the bus line number (from the database), 
 * the CDI line number (a CDI internal line number), the total number of registered devices, and the
 * parameter list (if any) supplied with the bus registration.
 *
 * \b example:
 * \include eg_busplug.c
 *
 * \return 0 upon success 
 */
CDI_EXPORT int cdiRegisterBusInitialization(char *busName,int (*fcn)(int,int,int,char *));

/**
 * \brief registers the bus handler dispatch function used for bus io.
 * 
 * This is the principal dispatch routine used to commuicate with the hardware bus.  All 
 * 'SEND' and 'RECV' activity flow through this routine.  Any calls will provide the complete
 * CdiRequestBlk defining the requested transaction, and it is the duty of the bus plug to process
 * the request and return.
 * 
 * \param busName is the bus plug name to be used by database entries refering to this bus plug.
 * \param fcn is the address of the bus handler function to be called when CDI accesses the hardware
 * bus. The function must have the prototype 
 * \n    int fcn(CdiRequestBlk *reqBlk)
 * \n and return '0' if the call was successful.  The CDI Request Block will define the nature of 
 * the requested transaction in its entirety.  All associated parameter lists will be provided within
 * this structure.
 *
 * \b example:
 * \include eg_busplug.c
 *
 * \return 0 upon success 
 */
CDI_EXPORT int cdiRegisterBusHandler(char *busName,void (*fcn)(CdiRequestInfoBlk *));

/**
 * \brief registers the bus cleanup function used when CDI closes.
 * 
 * This routine will be called when CDI closes and allows the bus to free up resources, etc.
 * 
 * \param busName is the bus plug name to be used by database entries refering to this bus plug.
 * \param fcn is the address of the bus cleanup function to be called when CDI closes.
 * The function must have the prototype 
 * \n    int fcn(int line)
 * \n and return '0' if the call was successful.  The line parameter will contain the bus line which
 * is being closed.
 *
 * \b example:
 * \include eg_busplug.c
 *
 * \return 0 upon success 
 */
CDI_EXPORT int cdiRegisterBusCleanup(char *busName,int (*fcn)(int));
CDI_EXPORT int cdiRegisterBus(char *busName,int busType);
CDI_EXPORT int cdiScanBus(char *busName,char *target,char *result,int resultBufferSize);

/**
 * \brief registers the bus scan function used for CDI diagnostics.
 * 
 * This routine will be called when CDI closes and allows the bus to free up resources, etc.
 * 
 * \param busName is the bus plug name to be used by database entries refering to this bus plug.
 * \param fcn is the address of the bus scan function to be called when CDI issues a bus scan.
 * It depends completely on the bus as to what consitutes a scan operation.
 * The function must have the prototype 
 * \n    int fcn(char *target,char *result,int resultBufferSize)
 * \n and return '0' if the call was successful.  The 'target' will simply be a character string
 * giving the desired scan operation (e.g. 'crates', 'addresses').  The 'result' will likewise
 * be a character string giving the results of the scan.  'resultBufferSize' gives the total
 * string capacity of 'result', so that the bus plug can know note to overwrite the given buffer.
 *
 * \b example:
 * \include eg_busplug.c
 *
 * \return 0 upon success 
 */
CDI_EXPORT int cdiRegisterBusScanner(char *busName,int (*fcn)(char *,char *,int));
CDI_EXPORT int cdiRegisterBusBlockFunction(char *busName,int (*fcn)(int *,int,void **,int));
CDI_EXPORT int cdiRegisterBusReqTimeout(char *busName,int (*fcn)(CdiRequestInfoBlk *, int, int));
CDI_EXPORT int cdiRegisterBusDeviceReady(char *busName,int (*fcn)(CdiRequestInfoBlk *, int));
CDI_EXPORT int cdiRegisterBusLoadRequest(char *busName,int (*fcn)(CdiRequestInfoBlk *, int));
CDI_EXPORT int cdiRegisterBusCancelRequest(char *busName,int (*fcn)(CdiRequestInfoBlk *, int));
CDI_EXPORT CdiLineInfoBlk *cdiGetBusHandle(int cdiLine);
CDI_EXPORT void cdi_Cleanup(void);
CDI_EXPORT void cdiCloseLinks(void);
CDI_EXPORT int cdi_assignDataLimits(char *devName,int dataSizeOut,int dataSizeIn,int dataSize2ndOut);
CDI_EXPORT int cdi_assignDefaultInput(char *devName,BYTE *pDataInput,int dataSize,int dataFormatIn);
CDI_EXPORT int cdi_CalibrateRequest(int requestIndex,int applyRules);
CDI_EXPORT int cdi_assignDataFormat(char *devName,int dataFormat);
CDI_EXPORT int cdi_assignDefaultInterval(char *devName,int interval);
CDI_EXPORT int cdi_getDefaultInterval(char *devName);
CDI_EXPORT int cdi_assignDataCalibration(char *devName,CdiRule *clbrRules,CdiRule *rvrsRules,int dataBusFormat);
CDI_EXPORT int getTypeLength(short dataType);
int clbrRequest(CdiRequestInfoBlk *pReq,int applyRules);
CDI_EXPORT int get_RequestBufferPointer(int requestIndex, CdiRequestInfoBlk **pReq);
CDI_EXPORT int cdi_getBusAddressParametersForDevice(char *devName,int *addrlist,char *addrmap);
CDI_EXPORT int cdi_getDeviceParametersForDevice(char *devName,int *paramlist);

#define CDI_INFO_ERRORS 1
#define CDI_INFO_NAMES  2
#define CDI_INFO_STATUS 3
#define CDI_INFO_NDEVS  4
#define CDI_INFO_ERRCNT 5

CDI_EXPORT int getTimeMilliseconds(void);
CDI_EXPORT int getGroupRequestInfo(int requestIndex, void *buffers, int index);
CDI_EXPORT int cdi_getTemplateNames(NAME16 *namebuffer,int namesize);
CDI_EXPORT int cdi_getTemplate(char *templateName,NAME16II *cdiTemplate,int *namesize);
CDI_EXPORT int cdi_getDescriptionForDevice(char *devName,char *devDesc);
CDI_EXPORT int assignGroupInOutFormatCode(CdiRequestInfoBlk *pReq, int inFormat, int outFormat);
CDI_EXPORT int cdi_getUnitsForDevice(char *devName,char *devUnits);
CDI_EXPORT int cdi_getPrecisionForDevice(char *devName,char *devPrec);
CDI_EXPORT int cdi_getRangeForDevice(char *devName,float *devRange);
CDI_EXPORT int cdi_getFormatForDevice(char *devName,int *devFmt);
CDI_EXPORT int cdi_getOutputSizeForDevice(char *devName,int *devSiz);
CDI_EXPORT int cdi_getLimitSizeForDevice(char *devName,int *devOutputSiz,int *devInputSiz);
CDI_EXPORT int cdi_RegisterCalibrationFunction(char *fcnName,double (*fcn)(double));
/**
 * \brief registers a calibration function for use by CDI
 * 
 * Bus actions 'SEND' and 'RECV' send and receive raw data to and from the
 * hardware bus according to the bus format registered with the corresponding device.
 * The raw data can be automatically calibrated (receive) or reverse calibrated (send)
 * by CDI if calibration rules are specified.  If the calibration rules are too complex
 * to be dealt with be the systematically known rules set, then one can make use of a 
 * calibration function registered via this call.
 * 
 * \param fcnName is the function name in plain text to be stored in the CDI function library.
 * \param fcn is the address of the function to be called when the calibation rule is to be used.
 * The function must take a one paramter as input (data type double) and return the calibrated
 * parameter (also as data type double).
 *
 * \b example:
 * \include eg_calbrfcn.c
 *
 * \return 0 upon success 
 */
CDI_EXPORT int cdiRegisterCalibrationFunction(char *fcnName,double (*fcn)(double));

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* CDI_H_ */
