/** \file cdiDbTools.c
 * \brief CDI-TINE API library routines
 *      
 */
#ifndef NO_DOXYGEN
#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#include "cdi.h"
#include "tinecdi.h"
#include "cdidb.h"
#endif /* NO_DOXYGEN */

/* synonyms: 
Address = Address Base
Rule = Rule Recv
Long Name = Address Map
*/

CdiDbEntry *cdiDbEntryLst = NULL;
int sizCdiEntryLst = 0;
int capCdiEntryLst = 0;

CdiMfDbEntry *cdiMfDbEntryLst = NULL;
int sizCdiMfEntryLst = 0;

CdiSchDbEntry *cdiSchDbEntryLst = NULL;
int sizCdiSchEntryLst = 0;

CdiSrvDbEntry *cdiSrvDbEntryLst = NULL;
int sizCdiSrvEntryLst = 0;

int registerDbStructs(void)
{
  int cc=0;
  static int done = 0;
  if (done) return 0;
  done = TRUE;
  /* this must follow the order of the structure explicitly! */
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,bus),132,CF_TEXT,"bus")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,name),32,CF_TEXT,"name")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,description),96,CF_TEXT,"desc")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,base),96,CF_TEXT,"addr")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,params),96,CF_TEXT,"params")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,map),64,CF_TEXT,"map")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,access),64,CF_TEXT,"access")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,line),1,CF_INT32,"line")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,number),1,CF_INT32,"number")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,interval),1,CF_INT32,"interval")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,mask),64,CF_TEXT,"mask")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,pattern),64,CF_TEXT,"pattern")) != 0) ccerr(cc);  
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,tolerance),16,CF_TEXT,"tolerance")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,dataformat),32,CF_TEXT,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,datainput),132,CF_TEXT,"input")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,datalimit),64,CF_TEXT,"limit")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,clbrrule_rcv),132,CF_TEXT,"rules_rcv")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(DBITEM_TAG,OFFSETIN(CdiDbEntry,clbrrule_snd),132,CF_TEXT,"rules_snd")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(DBITEM_TAG,sizeof(CdiDbEntry),DB_NICE_QUERY_SIZE)) != 0) ccerr(cc);

  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,bus),32,CF_TEXT,"bus")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,lib),64,CF_TEXT,"lib")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,env),128,CF_TEXT,"env")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,id),1,CF_INT32,"id")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,simulation),1,CF_INT32,"simulate")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,almThreshold),1,CF_INT32,"almThr")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(MFITEM_TAG,OFFSETIN(CdiMfDbEntry,reserved),1,CF_INT32,"reserved")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(MFITEM_TAG,sizeof(CdiMfDbEntry),MF_NICE_QUERY_SIZE)) != 0) ccerr(cc);

  if ((cc=addFieldToStruct(SCHDITEM_TAG,OFFSETIN(CdiSchDbEntry,prp),64,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SCHDITEM_TAG,OFFSETIN(CdiSchDbEntry,dev),64,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SCHDITEM_TAG,OFFSETIN(CdiSchDbEntry,tol),64,CF_TEXT,"tolerance")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(SCHDITEM_TAG,sizeof(CdiSchDbEntry),SCH_NICE_QUERY_SIZE)) != 0) ccerr(cc);

  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,prp),64,CF_TEXT,"property")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,dev),64,CF_TEXT,"device")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,fmt),16,CF_TEXT,"format")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,tol),64,CF_TEXT,"tolerance")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,siz),1,CF_INT32,"size")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,tmr),1,CF_INT32,"interval")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,hst),1,CF_INT32,"history")) != 0) ccerr(cc);
  if ((cc=addFieldToStruct(SRVITEM_TAG,OFFSETIN(CdiSrvDbEntry,reserved),1,CF_INT32,"reserved")) != 0) ccerr(cc);
  if ((cc=sealTaggedStruct(SRVITEM_TAG,sizeof(CdiSrvDbEntry),SRV_NICE_QUERY_SIZE)) != 0) ccerr(cc);

err:
  return cc;
}
int getMfDbItemIndex(CdiMfDbEntry *item)
{
  int i;
  for (i=0; i<sizCdiMfEntryLst; i++)
  {
    if (strnicmp(cdiMfDbEntryLst[i].bus,item->bus,32)) continue;
    return i;
  }
  return -1;
}
int getSrvDbItemIndex(CdiSrvDbEntry *item)
{
  int i;
  for (i=0; i<sizCdiSrvEntryLst; i++)
  {
    if (strnicmp(cdiSrvDbEntryLst[i].prp,item->prp,64)) continue;
    if (strnicmp(cdiSrvDbEntryLst[i].dev,item->dev,64)) continue;
    return i;
  }
  return -1;
}
int getSchDbItemIndex(CdiSchDbEntry *item)
{
  int i;
  for (i=0; i<sizCdiSchEntryLst; i++)
  {
    if (strnicmp(cdiSchDbEntryLst[i].prp,item->prp,64)) continue;
    if (strnicmp(cdiSchDbEntryLst[i].dev,item->dev,64)) continue;
    return i;
  }
  return -1;
}
int getDbItemIndex(CdiDbEntry *item,int matchBusName)
{ /* internal: item never NULL here */
  int i,busIdx=-1,busNameFound=FALSE;
  for (i=0; i<sizCdiEntryLst; i++)
  {
    if (!matchBusName)
    {
      if (strnicmp(cdiDbEntryLst[i].name,item->name,32)) continue;
      return i;
    }
    else
    {
      if (!busNameFound)
      {  
        if (strnicmp(cdiDbEntryLst[i].bus,item->bus,132)) continue;
        busNameFound = TRUE;
      }
      if (strnicmp(cdiDbEntryLst[i].bus,item->bus,132)) return busIdx;
      busIdx = i;
    }
  }
  return busIdx;
}
int isValidString(const char *tag,const char *key,int len,const char *invSet,int emptyStringAllowed)
{ /* internal: key, invSet never NULL ! */
  char *fld,fldBuffer[256];
  int slen = (int)strlen(key);
  if (slen == 0 && !emptyStringAllowed) 
  {
    srvlog(CDIDBLOGFILE,"isValidString","supplied string for %s is not allowed to be empty!",tag);
    return FALSE;
  }
  if (invSet == NULL) return TRUE;
  if (slen < len) 
  {
    fld = (char *)key;
  }
  else
  {
    strncpy(fldBuffer,key,len);
    fldBuffer[len] = 0;
    fld = fldBuffer;
  }
  if (strpbrk(fld,invSet) == NULL) return TRUE;
  srvlog(CDIDBLOGFILE,"isValidString","supplied string for %s (%s) contains an invalid character!",tag,fld);
  return FALSE;
}
static char *invNameSet = ", ;'\"\\~*^(){}[]?<>";
static char *invAddrSet = ", ;'\"\\~*^(){}[]?";
static char *invPathSet = ", ;'\"~*^(){}[]?";
int isValidDbItem(CdiDbEntry *item)
{
  if (item == NULL) return FALSE;

  if (!isValidString("BUS",item->bus,132,invNameSet,FALSE)) return FALSE;
  if (!isValidString("NAME",item->name,32,invNameSet,FALSE)) return FALSE;
  if (!isValidString("ADDRESS BASE",item->base,96,invAddrSet,TRUE)) return FALSE;
  if (!isValidString("ADDRESS PARAMETERS",item->params,96,invAddrSet,TRUE)) return FALSE;
  if (!isValidString("ADDRESS MAP",item->map,64,NULL,TRUE)) return FALSE;
  if (!isValidString("ACCESS",item->access,64,NULL,TRUE)) return FALSE;
  if (!isValidString("MASK",item->mask,64,NULL,TRUE)) return FALSE;
  if (!isValidString("PATTERN",item->pattern,64,NULL,TRUE)) return FALSE;
  if (!isValidString("TOLERANCE",item->tolerance,16,NULL,TRUE)) return FALSE;
  if (!isValidString("FORMAT",item->dataformat,32,NULL,TRUE)) return FALSE;
  if (!isValidString("INPUT",item->datainput,132,NULL,TRUE)) return FALSE;
  if (!isValidString("LIMIT",item->datalimit,64,NULL,TRUE)) return FALSE;
  if (!isValidString("RULE_RCV",item->clbrrule_rcv,132,NULL,TRUE)) return FALSE;
  if (!isValidString("RULE_SND",item->clbrrule_snd,132,NULL,TRUE)) return FALSE;
  return TRUE;
}
int isValidMfDbItem(CdiMfDbEntry *item)
{
  if (item == NULL) return FALSE;
  if (!isValidString("BUS",item->bus,132,invNameSet,FALSE)) return FALSE;
  if (!isValidString("LIBRARY",item->lib,32,invPathSet,FALSE)) return FALSE;
  return TRUE;
}
int isValidSrvDbItem(CdiSrvDbEntry *item)
{
  if (item == NULL) return FALSE;
  if (!isValidString("PROPERTY",item->prp,64,invNameSet,FALSE)) return FALSE;
  if (!isValidString("DEVICE",item->dev,64,invNameSet,FALSE)) return FALSE;
  return TRUE;
}
int isValidSchDbItem(CdiSchDbEntry *item)
{
  if (item == NULL) return FALSE;
  if (!isValidString("PROPERTY",item->prp,64,invNameSet,FALSE)) return FALSE;
  if (!isValidString("DEVICE",item->dev,64,invNameSet,TRUE)) return FALSE;
  return TRUE;
}
int addrdbcmp(CdiDbEntry *entries,int len)
{
  if (len != sizCdiEntryLst) return TRUE; /* obviously not identical */
  return memcmp(entries,cdiDbEntryLst,len*sizeof(CdiDbEntry));
}
int useDbItemsAsAddrDb(CdiDbEntry *entries,int len)
{
  if (addrdbcmp(entries,len) == 0) return already_assigned; /* no need to do anything */
  if (cdiDbEntryLst != NULL) SystemFree(cdiDbEntryLst);
  cdiDbEntryLst = entries;
  sizCdiEntryLst = len;
  dumpDbItemList(NULL);
  cdiDbEntryLst = NULL;
  sizCdiEntryLst = 0;
  return 0;
}
int mfdbcmp(CdiMfDbEntry *entries,int len)
{
  if (len != sizCdiMfEntryLst) return TRUE; /* obviously not identical */
  return memcmp(entries,cdiMfDbEntryLst,len*sizeof(CdiMfDbEntry));
}
int useDbItemsAsMfDb(CdiMfDbEntry *entries,int len)
{
  if (mfdbcmp(entries,len) == 0) return already_assigned; /* no need to do anything */
  if (cdiMfDbEntryLst != NULL) SystemFree(cdiMfDbEntryLst);
  cdiMfDbEntryLst = entries;
  sizCdiMfEntryLst = len;
  dumpMfDbItemList(NULL);
  cdiMfDbEntryLst = NULL;
  sizCdiMfEntryLst = 0;
  return 0;
}
int srvdbcmp(CdiSrvDbEntry *entries,int len)
{
  if (len != sizCdiSrvEntryLst) return TRUE; /* obviously not identical */
  return memcmp(entries,cdiSrvDbEntryLst,len*sizeof(CdiSrvDbEntry));
}
int useDbItemsAsSrvDb(CdiSrvDbEntry *entries,int len)
{
  if (srvdbcmp(entries,len) == 0) return already_assigned; /* no need to do anything */
  if (cdiSrvDbEntryLst != NULL) SystemFree(cdiSrvDbEntryLst);
  cdiSrvDbEntryLst = entries;
  sizCdiSrvEntryLst = len;
  dumpSrvDbItemList(NULL);
  cdiSrvDbEntryLst = NULL;
  sizCdiSrvEntryLst = 0;
  return 0;
}
int schdbcmp(CdiSchDbEntry *entries,int len)
{
  if (len != sizCdiSchEntryLst) return TRUE; /* obviously not identical */
  return memcmp(entries,cdiSchDbEntryLst,len*sizeof(CdiSchDbEntry));
}
int useDbItemsAsSchDb(CdiSchDbEntry *entries,int len)
{
  if (schdbcmp(entries,len) == 0) return already_assigned; /* no need to do anything */
  if (cdiSchDbEntryLst != NULL) SystemFree(cdiSchDbEntryLst);
  cdiSchDbEntryLst = entries;
  sizCdiSchEntryLst = len;
  dumpSchDbItemList(NULL);
  cdiSchDbEntryLst = NULL;
  sizCdiSchEntryLst = 0;
  return 0;
}
int filecopy(char *dstfn,char *srcfn)
{
  FILE *dfp, *sfp;
  char str[256];
  int cc = 0;
  if ((sfp=fopen(srcfn,"rt")) == NULL) ccerr(file_error);
  if ((dfp=fopen(dstfn,"wt")) == NULL) ccerr(file_error);

  while (fgets(str,255,sfp) != NULL) /* scan file */
  {
    fputs(str,dfp);
  }
err:
  if (dfp != NULL) fclose(dfp);
  if (sfp != NULL) fclose(sfp);
  return cc;
}
int rollbackMfItems(void)
{
  char fname[128];
  if (strlen(cdiDbRb.mf) == 0) return no_such_file;
  if (cdiMfDbEntryLst != NULL) SystemFree(cdiMfDbEntryLst);
  cdiMfDbEntryLst = NULL;
  sizCdiMfEntryLst = 0;
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDIMF_FILE);
  if (filecopy(fname,cdiDbRb.mf) != 0)
  {
    cdilog("could not rollback cdi manifest file !");
    return file_error;
  }
  return 0;
}
int rollbackSrvItems(void)
{
  char fname[128];
  if (strlen(cdiDbRb.srv) == 0) return no_such_file;
  if (cdiSrvDbEntryLst != NULL) SystemFree(cdiSrvDbEntryLst);
  cdiSrvDbEntryLst = NULL;
  sizCdiSrvEntryLst = 0;
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDISRV_FILE);
  if (filecopy(fname,cdiDbRb.srv) != 0)
  {
    cdilog("could not rollback cdi server db file !");
    return file_error;
  }
  return 0;
}
int rollbackSchItems(void)
{
  char fname[128];
  if (strlen(cdiDbRb.sch) == 0) return no_such_file;
  if (cdiSchDbEntryLst != NULL) SystemFree(cdiSchDbEntryLst);
  cdiSchDbEntryLst = NULL;
  sizCdiSchEntryLst = 0;
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDISCHED_FILE);
  if (filecopy(fname,cdiDbRb.sch) != 0)
  {
    cdilog("could not rollback cdi schedule file !");
    return file_error;
  }
  return 0;
}
int rollbackAddrItems(void)
{
  char fname[128];
  if (strlen(cdiDbRb.addr) == 0) return no_such_file;
  if (cdiDbEntryLst != NULL) SystemFree(cdiDbEntryLst);
  cdiDbEntryLst = NULL;
  sizCdiEntryLst = 0;
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDIDB_FILE);
  if (filecopy(fname,cdiDbRb.addr) != 0)
  {
    cdilog("could not rollback cdi address file !");
    return file_error;
  }
  return 0;
}

DbMergeItem *dbMergeLst = NULL;

void freeMergeList(void)
{
  DbMergeItem *mrg, *nxt;
  for (mrg=dbMergeLst; mrg != NULL; mrg=nxt)
  {
    nxt = mrg->nxt;
    SystemFree(mrg);
  }
  dbMergeLst = NULL;
}
void dbItemSpice(DbMergeItem *mrg1,DbMergeItem *mrg2)
{ /* mrg1 is farther along then mrg2 and needs to be inserted prior to mrg2 */
  if (mrg1 == NULL || mrg2 == NULL) return;
  if (mrg1->prv != NULL) mrg1->prv->nxt = mrg1->nxt; /* mrg1->prv should never be NULL */
  mrg1->prv = mrg2->prv;
  mrg1->nxt = mrg2;
  mrg2->prv = mrg1;
}
int isBusAtStart(char *bus)
{
  if (!stricmp(bus,"TEMPLATE")) return TRUE;
  if (!stricmp(bus,"BITFIELD")) return TRUE;
  if (!stricmp(bus,"FIELDBUS")) return TRUE;
  return FALSE;
}
int mergeDbItems(CdiDbEntry *entries,int len)
{
  int cc=0,i,k,n=0,idx=-1,need_sorted=FALSE,last_idx=-1;
  CdiDbEntry *item;
  DbMergeItem *mrg, *idxmrg, *tail=NULL;
  if (entries == NULL) return invalid_parameter;
  if (len < 1) return dimension_error;
  populateDbItemList();
  for (i=0; i<len; i++)
  {
    item = &entries[i];
    if (!isValidDbItem(item)) continue;
    if ((k=getDbItemIndex(item,FALSE)) >= 0)
    { /* this is an edit ... */
      memcpy(&cdiDbEntryLst[k],item,sizeof(CdiDbEntry));
      idx = k;
      continue;
    }
    if ((k=getDbItemIndex(item,TRUE)) >= 0) idx = k;
    if ((mrg=(DbMergeItem *)SystemCalloc(1,sizeof(DbMergeItem))) == NULL) ccerr(out_of_server_memory);
    if (!isBusAtStart(item->bus))
    { /* not found will go at the end of the database */
      mrg->idx = idx < 0 ? (sizCdiEntryLst + n) : idx;
    }
    else
    { /* not found will go at the beginning of the database */
      mrg->idx = idx;
    }
    mrg->item = item;
    if (mrg->idx < last_idx) need_sorted = TRUE;
    last_idx = mrg->idx;
    /* make this a forward linked list (increasing 'idx') */
    if (tail != NULL) {tail->nxt = mrg; mrg->prv = tail;} else dbMergeLst = mrg;
    tail = mrg;
    n++;
  }
  if (need_sorted)
  { /* added entries are not in order */
    idx = -1; idxmrg = NULL;
    for (mrg=dbMergeLst; mrg != NULL; mrg=mrg->nxt)
    {
      if (mrg->idx < idx) 
      {
        tail = mrg->prv;
        dbItemSpice(mrg,idxmrg); /* put mrg prior to idxmrg */
        mrg = tail; /* so that the iteration can continue */
        continue;
      }
      idx = mrg->idx;
      idxmrg = mrg;
    }
  }
  cc = dumpDbItemList(dbMergeLst);
  freeMergeList();
err:
  return cc;
}
int mergeMfDbItems(CdiMfDbEntry *entries,int len)
{
  int cc=0,i,k;
  CdiMfDbEntry *item;
  DbMergeItem *mrg, *tail=NULL;
  if (entries == NULL) return invalid_parameter;
  if (len < 1) return dimension_error;
  populateMfItemList();
  for (i=0; i<len; i++)
  {
    item = &entries[i];
    if (!isValidMfDbItem(item)) continue;
    if ((k=getMfDbItemIndex(item)) >= 0)
    { /* this is an edit ... */
      memcpy(&cdiDbEntryLst[k],item,sizeof(CdiDbEntry));
      continue;
    }
    if ((mrg=(DbMergeItem *)SystemCalloc(1,sizeof(DbMergeItem))) == NULL) ccerr(out_of_server_memory);
    mrg->item = item;
    if (tail != NULL) {tail->nxt = mrg; mrg->prv = tail;} else dbMergeLst = mrg;
    tail = mrg;
  }
  cc = dumpMfDbItemList(dbMergeLst);
  freeMergeList();
err:
  return cc;
}
int mergeSrvDbItems(CdiSrvDbEntry *entries,int len)
{
  int cc=0,i,k;
  CdiSrvDbEntry *item;
  DbMergeItem *mrg, *tail=NULL;
  if (entries == NULL) return invalid_parameter;
  if (len < 1) return dimension_error;
  populateSrvItemList();
  for (i=0; i<len; i++)
  {
    item = &entries[i];
    if (!isValidSrvDbItem(item)) continue;
    if ((k=getSrvDbItemIndex(item)) >= 0)
    { /* this is an edit ... */
      memcpy(&cdiSrvDbEntryLst[k],item,sizeof(CdiSrvDbEntry));
      continue;
    }
    if ((mrg=(DbMergeItem *)SystemCalloc(1,sizeof(DbMergeItem))) == NULL) ccerr(out_of_server_memory);
    mrg->item = item;
    if (tail != NULL) {tail->nxt = mrg; mrg->prv = tail;} else dbMergeLst = mrg;
    tail = mrg;
  }
  cc = dumpSrvDbItemList(dbMergeLst);
  freeMergeList();
err:
  return cc;
}
int mergeSchDbItems(CdiSchDbEntry *entries,int len)
{
  int cc=0,i,k;
  CdiSchDbEntry *item;
  DbMergeItem *mrg, *tail=NULL;
  if (entries == NULL) return invalid_parameter;
  if (len < 1) return dimension_error;
  populateSchItemList();
  for (i=0; i<len; i++)
  {
    item = &entries[i];
    if (!isValidSchDbItem(item)) continue;
    if ((k=getSchDbItemIndex(item)) >= 0)
    { /* this is an edit ... */
      memcpy(&cdiSchDbEntryLst[k],item,sizeof(CdiSchDbEntry));
      continue;
    }
    if ((mrg=(DbMergeItem *)SystemCalloc(1,sizeof(DbMergeItem))) == NULL) ccerr(out_of_server_memory);
    mrg->item = item;
    if (tail != NULL) {tail->nxt = mrg; mrg->prv = tail;} else dbMergeLst = mrg;
    tail = mrg;
  }
  cc = dumpSchDbItemList(dbMergeLst);
  freeMergeList();
err:
  return cc;
}
void writeDbItemToFile(FILE *fp,CdiDbEntry *lst)
{
  char numStr[32],ivlStr[32];
  if (lst->number > 0) sprintf(numStr,"%d",lst->number); else numStr[0] = 0;
  if (lst->interval > 0) sprintf(ivlStr,"%d",lst->interval); else ivlStr[0] = 0;
  fprintf(fp,"%.32s, %.132s, %.96s, %.96s, %.96s, %.64s, %.64s, %d, %.16s, %.16s, %.64s, %.64s, %.16s, %.32s, %.132s, %.64s, %.132s, %.132s\n",
    lst->name,lst->bus,lst->description,lst->base,lst->params,lst->map,lst->access,
    lst->line,numStr,ivlStr,lst->mask,lst->pattern,lst->tolerance,
    lst->dataformat,lst->datainput,lst->datalimit,lst->clbrrule_rcv,lst->clbrrule_snd);
  return;
}
int cdiDbFileExists(char *fname)
{
  struct stat sbuf;
  if (stat(fname,&sbuf )) return FALSE;
  return TRUE;
}
int dumpDbItemList(DbMergeItem *mrg)
{
  FILE *fp = NULL;
  char fname_bak[128],fname[128],bakDecor[32];
  time_t t = time(NULL);
  struct tm *tms;
  int cc = 0, i, n=0, advance=FALSE;
  CdiDbEntry *lst;

  if ((tms=localtime(&t)) == NULL) return code_failure;
  /* backup current database */
  sprintf(bakDecor,"-%02d.%02d.%02d-%05d",tms->tm_mday,tms->tm_mon+1,tms->tm_year%100,(int)(t%86400));
  sprintf(fname_bak,"%.96s%.16s%.30s",CdiDBpath,CDIDB_FILE,bakDecor);
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDIDB_FILE);
  if (cdiDbFileExists(fname))
  { /* the db file exists */
    if (rename(fname,fname_bak) != 0) 
    {
      cdilog("could not backup cdi address file !");
      return file_error;
    }
    strncpy(cdiDbRb.addr,fname_bak,128); /* keep track of rollback db names */
    putDbRbList();
    cdilog("backed up address file to %s",fname_bak);
  }
  if ((fp=fopen(fname,"w")) == NULL)
  {
    cdilog("could not open cdi address file for writing !");
    return file_error;
  }
  fprintf(fp,"%s\n",CdiDbHdr);
  for (i=0; i<sizCdiEntryLst; i++)
  {
    lst = &cdiDbEntryLst[i];
    if (!stricmp(lst->bus,"REMOVE")) continue;
merge_loop:
    if (mrg != NULL && mrg->idx < i)
    {
      lst = (CdiDbEntry *)mrg->item;
      advance = TRUE;
    }
    writeDbItemToFile(fp,lst);
    n++;
    if (mrg != NULL && advance)
    {
      mrg = mrg->nxt;
      advance = FALSE;
      goto merge_loop;
    }
  }
  while (mrg != NULL)
  { /* anything left to write ? */
    writeDbItemToFile(fp,(CdiDbEntry *)mrg->item);
    n++;
    mrg = mrg->nxt;
  }
  fclose(fp);
  cdilog("wrote new cdi address file (%d entries)",n);
  return cc;
}
/* local copy of csvStdInit seems to be necessary (heap problem?) */
void cdiStdInit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  db->cnt = 0;
  if (db->tgt == NULL) db->tgt = (BYTE *)SystemCalloc(db->siz,db->tgt_siz);
  for (i=0; i<db->ncols; i++) 
  {
    db->col[i].start = db->col[i].ptr; /* reset via csvStdExit() */
    db->col[i].ptr = db->tgt + (size_t)db->col[i].start; 
  }
}
CSVCOLUMN csvCdiSchDbColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSchDbEntry,prp),NULL,0 },
  { "DEVICE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSchDbEntry,dev),NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSchDbEntry,tol),NULL,0 },
};
CSVDB csvCdiSchDb =
{
  csvCdiSchDbColumns,
  sizeof(csvCdiSchDbColumns)/sizeof(csvCdiSchDbColumns[0]),
  0,
  cdiStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CdiSchDbEntry),
  0
};
int populateSchItemList(void)
{
  int cc;
  if (cdiSchDbEntryLst != NULL) return already_assigned;
  if ((cc=csvReadFile(CdiDBpath,CDISCHED_FILE,&csvCdiSchDb,(void **)&cdiSchDbEntryLst)) != 0) return cc;
  sizCdiSchEntryLst = csvCdiSchDb.siz;
  return 0;
}

CSVCOLUMN csvCdiMfDbColumns[] =
{
  { "BUS","",0,CF_TEXT,32,(BYTE *)OFFSETIN(CdiMfDbEntry,bus),NULL,0 },
  { "LIBRARY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiMfDbEntry,lib),NULL,0 },
  { "BUS_ENV","",0,CF_TEXT,128,(BYTE *)OFFSETIN(CdiMfDbEntry,env),NULL,0 },
  { "BUS_ID","",0,CF_INT32,1,(BYTE *)OFFSETIN(CdiMfDbEntry,id),NULL,0 },
  { "SIMULATION","",0,CF_INT32,1,(BYTE *)OFFSETIN(CdiMfDbEntry,simulation),NULL,0 },
  { "ALARM_THRESHOLD","0",0,CF_INT32,1,(BYTE *)OFFSETIN(CdiMfDbEntry,almThreshold),NULL,0 },
};
CSVDB csvCdiMfLstDb =
{
  csvCdiMfDbColumns,
  sizeof(csvCdiMfDbColumns)/sizeof(csvCdiMfDbColumns[0]),
  0,
  cdiStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CdiMfDbEntry),
  0
};
int populateMfItemList(void)
{
  int cc;
  if (cdiMfDbEntryLst != NULL) return already_assigned;
  if ((cc=csvReadFile(CdiDBpath,CDIMF_FILE,&csvCdiMfLstDb,(void **)&cdiMfDbEntryLst)) != 0) return cc;
  sizCdiMfEntryLst = csvCdiMfLstDb.siz;
  return 0;
}
void writeMfDbItemToFile(FILE *fp,CdiMfDbEntry *lst)
{
  char idStr[32],simStr[32],thrStr[32];
  if (lst->id > 0) sprintf(idStr,"%d",lst->id); else idStr[0] = 0;
  if (lst->simulation > 0) sprintf(simStr,"%d",lst->simulation); else simStr[0] = 0;
  if (lst->almThreshold > 0) sprintf(thrStr,"%d",lst->almThreshold); else thrStr[0] = 0;
  fprintf(fp,"%.32s, %.64s, %.128s, %.32s, %.32s, %.32s\n",
    lst->bus,lst->lib,lst->env,idStr,simStr,thrStr);
  return;
}
int dumpMfDbItemList(DbMergeItem *mrg)
{
  FILE *fp = NULL;
  char fname_bak[128],fname[128],bakDecor[32];
  time_t t = time(NULL);
  struct tm *tms;
  int cc = 0, i, n=0;
  CdiMfDbEntry *lst;

  if ((tms=localtime(&t)) == NULL) return code_failure;
  /* backup current database */
  sprintf(bakDecor,"-%02d.%02d.%02d-%05d",tms->tm_mday,tms->tm_mon+1,tms->tm_year%100,(int)(t%86400));
  sprintf(fname_bak,"%.96s%.16s%.30s",CdiDBpath,CDIMF_FILE,bakDecor);
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDIMF_FILE);
  if (cdiDbFileExists(fname))
  { /* the db file exists */
    if (rename(fname,fname_bak) != 0) 
    {
      cdilog("could not backup cdi manifest file !");
      return file_error;
    }
    strncpy(cdiDbRb.mf,fname_bak,128); /* keep track of rollback db names */
    putDbRbList();
    cdilog("backed up manifest file to %s",fname_bak);
  }
  if ((fp=fopen(fname,"w")) == NULL)
  {
    cdilog("could not open cdi manifest file for writing !");
    return file_error;
  }
  fprintf(fp,"%s\n",CdiMfDbHdr);
  for (i=0; i<sizCdiMfEntryLst; i++)
  {
    lst = &cdiMfDbEntryLst[i];
    if (!stricmp(lst->bus,"REMOVE")) continue;
    writeMfDbItemToFile(fp,lst);
    n++;
  }
  while (mrg != NULL)
  { /* anything left to write ? */
    writeMfDbItemToFile(fp,(CdiMfDbEntry *)mrg->item);
    n++;
    mrg = mrg->nxt;
  }
  fclose(fp);
  cdilog("wrote new cdi manifest file (%d entries)",n);
  return cc;
}

CSVCOLUMN csvCdiSrvItemLstColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(CdiSrvDbEntry,prp),NULL,0 },
  { "DEVICE","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(CdiSrvDbEntry,dev),NULL,0 },
  { "FORMAT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(CdiSrvDbEntry,fmt),NULL,0 },
  { "SCHEDULE_TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSrvDbEntry,tol),NULL,0 },
  { "SIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(CdiSrvDbEntry,siz),NULL,0 },
  { "INTERVAL","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(CdiSrvDbEntry,tmr),NULL,0 },
  { "KEEP_HISTORY","0",0,CF_BOOLEAN,1,(BYTE *)OFFSETIN(CdiSrvDbEntry,hst),NULL,0 },
};
CSVDB csvCdiSrvItemLstDb =
{
  csvCdiSrvItemLstColumns,
  sizeof(csvCdiSrvItemLstColumns)/sizeof(csvCdiSrvItemLstColumns[0]),
  0,
  cdiStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CdiSrvDbEntry),
  0
};
int populateSrvItemList(void)
{
  int cc;
  if (cdiSrvDbEntryLst != NULL) return already_assigned;
  if ((cc=csvReadFile(CdiDBpath,CDISRV_FILE,&csvCdiSrvItemLstDb,(void **)&cdiSrvDbEntryLst)) != 0) return cc;
  sizCdiSrvEntryLst = csvCdiSrvItemLstDb.siz;

  return 0;
}
void writeSrvDbItemToFile(FILE *fp,CdiSrvDbEntry *lst)
{
  char sizStr[32],ivlStr[32],hstStr[32];
  if (lst->siz > 0) sprintf(sizStr,"%d",lst->siz); else sizStr[0] = 0;
  if (lst->tmr > 0) sprintf(ivlStr,"%d",lst->tmr); else ivlStr[0] = 0;
  if (lst->hst > 0) sprintf(hstStr,"%s",(lst->hst ? "TRUE" : "FALSE")); else hstStr[0] = 0;
  fprintf(fp,"%.64s, %.64s, %.16s, %.64s, %.32s, %.32s, %.32s\n",
    lst->prp,lst->dev,lst->fmt,lst->tol,sizStr,ivlStr,hstStr);
  return;
}
int dumpSrvDbItemList(DbMergeItem *mrg)
{
  FILE *fp = NULL;
  char fname_bak[128],fname[128],bakDecor[32];
  time_t t = time(NULL);
  struct tm *tms;
  int cc = 0, i, n=0;
  CdiSrvDbEntry *lst;

  if ((tms=localtime(&t)) == NULL) return code_failure;
  /* backup current database */
  sprintf(bakDecor,"-%02d.%02d.%02d-%05d",tms->tm_mday,tms->tm_mon+1,tms->tm_year%100,(int)(t%86400));
  sprintf(fname_bak,"%.96s%.16s%.30s",CdiDBpath,CDISRV_FILE,bakDecor);
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDISRV_FILE);
  if (cdiDbFileExists(fname))
  { /* the db file exists */
    if (rename(fname,fname_bak) != 0) 
    {
      cdilog("could not backup cdi server file !");
      return file_error;
    }
    strncpy(cdiDbRb.srv,fname_bak,128); /* keep track of rollback db names */
    putDbRbList();
    cdilog("backed up server file to %s",fname_bak);
  }
  if ((fp=fopen(fname,"w")) == NULL)
  {
    cdilog("could not open cdi server db file for writing !");
    return file_error;
  }
  fprintf(fp,"%s\n",CdiSrvDbHdr);
  for (i=0; i<sizCdiSrvEntryLst; i++)
  {
    lst = &cdiSrvDbEntryLst[i];
    if (!stricmp(lst->prp,"REMOVE")) continue;
    writeSrvDbItemToFile(fp,lst);
    n++;
  }
  while (mrg != NULL)
  { /* anything left to write ? */
    writeSrvDbItemToFile(fp,(CdiSrvDbEntry *)mrg->item);
    n++;
    mrg = mrg->nxt;
  }
  fclose(fp);
  cdilog("wrote new cdi server db file (%d entries)",n);
  return cc;
}
void writeSchDbItemToFile(FILE *fp,CdiSchDbEntry *lst)
{
  fprintf(fp,"%.64s, %.64s, %.64s\n",lst->prp,lst->dev,lst->tol);
  return;
}
int dumpSchDbItemList(DbMergeItem *mrg)
{
  FILE *fp = NULL;
  char fname_bak[128],fname[128],bakDecor[32];
  time_t t = time(NULL);
  struct tm *tms;
  int cc = 0, i, n=0;
  CdiSchDbEntry *lst;

  if ((tms=localtime(&t)) == NULL) return code_failure;
  /* backup current database */
  sprintf(bakDecor,"-%02d.%02d.%02d-%05d",tms->tm_mday,tms->tm_mon+1,tms->tm_year%100,(int)(t%86400));
  sprintf(fname_bak,"%.96s%.16s%.30s",CdiDBpath,CDISCHED_FILE,bakDecor);
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDISCHED_FILE);
  if (cdiDbFileExists(fname))
  { /* the db file exists */
    if (rename(fname,fname_bak) != 0) 
    {
      cdilog("could not backup cdi schedule file !");
      return file_error;
    }
    strncpy(cdiDbRb.sch,fname_bak,128); /* keep track of rollback db names */
    putDbRbList();
    cdilog("backed up schedule file to %s",fname_bak);
  }
  if ((fp=fopen(fname,"w")) == NULL)
  {
    cdilog("could not open cdi schedule db file for writing !");
    return file_error;
  }
  fprintf(fp,"%s\n",CdiSchDbHdr);
  for (i=0; i<sizCdiSchEntryLst; i++)
  {
    lst = &cdiSchDbEntryLst[i];
    if (!stricmp(lst->prp,"REMOVE")) continue;
    writeSchDbItemToFile(fp,lst);
    n++;
  }
  while (mrg != NULL)
  { /* anything left to write ? */
    writeSchDbItemToFile(fp,(CdiSchDbEntry *)mrg->item);
    n++;
    mrg = mrg->nxt;
  }
  fclose(fp);
  cdilog("wrote new cdi schedule db file (%d entries)",n);
  return cc;
}
int populateDbItemList(void)
{
  char *c,str[256],scratch[132],fname[132];
  int done = 0,i,n,siz, cc = 0;
  int bus_col=-1,lin_col=-1,addr_col=-1,bas_col=-1,prm_col=-1,msk_col=-1;
  int pat_col=-1,acc_col=-1,tol_col=-1,num_col=-1,ivl_col=-1,nam_col=-1;
  int fmt_col=-1,ipt_col=-1,lim_col=-1,rul_col=-1,rrcv_col=-1,rsnd_col=-1;
  int lng_col=-1,map_col=-1,dsc_col=-1;
  FILE *fp = NULL;
  CdiDbEntry *lst = NULL;

  if (cdiDbEntryLst != NULL)
  { /* already populated */
    return already_assigned;
  }
  sprintf(fname,"%.96s%.16s",CdiDBpath,CDIDB_FILE); 
  if ((siz=GetCsvDataBaseSize(fname)) < 0) return -siz;
  n = siz + DB_OVERFLOW;
  if ((lst=(CdiDbEntry *)SystemCalloc(n,sizeof(CdiDbEntry))) == NULL) return out_of_server_memory;
  if (siz == 0) 
  {
    cdiDbEntryLst = lst;
    sizCdiEntryLst = 0;
    capCdiEntryLst = n;
    ccerr(0);
  }
  if ((fp=fopen(fname,"rt")) == NULL) ccerr(no_such_file);
  for (i=0; fgets(str,255,fp) != NULL && i<n; ) /* scan file */
  {
    strtrm(str);
    if (strlen(str) == 0) continue; /* carriage return */
    if (strchr("%;#\n",str[0])) continue; /* comment line(s) */
    if (!done)
    {
      bus_col = GetCsvColumnIndex(str,"BUS");
      lin_col = GetCsvColumnIndex(str,"LINE");
      addr_col = GetCsvColumnIndex(str,"ADDRESS");
      bas_col = GetCsvColumnIndex(str,"ADDRESS_BASE");
      prm_col = GetCsvColumnIndex(str,"ADDRESS_PARAMETERS");
      msk_col = GetCsvColumnIndex(str,"MASK");
      pat_col = GetCsvColumnIndex(str,"PATTERN");
      acc_col = GetCsvColumnIndex(str,"ACCESS");
      tol_col = GetCsvColumnIndex(str,"TOLERANCE");
      num_col = GetCsvColumnIndex(str,"NUMBER"); /* 0 => auto increment */
      ivl_col = GetCsvColumnIndex(str,"INTERVAL"); /* default = 1000 */
      nam_col = GetCsvColumnIndex(str,"NAME");
      fmt_col = GetCsvColumnIndex(str,"FORMAT"); /* default = INT16 */
      ipt_col = GetCsvColumnIndex(str,"INPUT");
      lim_col = GetCsvColumnIndex(str,"LIMIT");
      rul_col = GetCsvColumnIndex(str,"RULE");
      rrcv_col = GetCsvColumnIndex(str,"RULE_RECV");
      rsnd_col = GetCsvColumnIndex(str,"RULE_SEND");
      lng_col = GetCsvColumnIndex(str,"LONG_NAME");
      map_col = GetCsvColumnIndex(str,"ADDRESS_MAP");
      dsc_col = GetCsvColumnIndex(str,"DESCRIPTION");
      done = 1;
      continue;
    }

    if (bus_col >= 0) strncpy(lst[i].bus,GetCsvColumnString(bus_col,str),132);
    if (lin_col >= 0)
    {
      strncpy(scratch,GetCsvColumnString(lin_col,str),64);
      lst[i].line = atoi(scratch);
    }
    if (addr_col >= 0)
    {
      strncpy(scratch,GetCsvColumnString(addr_col,str),132);
      if ((c=strchr(scratch,':')) != NULL)
      {
        *c++ = 0;
        strncpy(lst[i].base,scratch,96);
        strncpy(lst[i].params,c,96);
      }
    }
    if (bas_col >= 0) strncpy(lst[i].base,GetCsvColumnString(bas_col,str),96);
    if (prm_col >= 0) strncpy(lst[i].params,GetCsvColumnString(prm_col,str),96);    
    if (msk_col >= 0) strncpy(lst[i].mask,GetCsvColumnString(msk_col,str),64);
    if (pat_col >= 0) strncpy(lst[i].pattern,GetCsvColumnString(pat_col,str),64);
    if (acc_col >= 0) strncpy(lst[i].access,GetCsvColumnString(acc_col,str),64);
    if (tol_col >= 0) strncpy(lst[i].tolerance,GetCsvColumnString(tol_col,str),64);
    if (num_col >= 0) 
    {
      strncpy(scratch,GetCsvColumnString(num_col,str),64);
      lst[i].number = atoi(scratch);
    }
    if (ivl_col >= 0) 
    {
      strncpy(scratch,GetCsvColumnString(ivl_col,str),64);
      lst[i].interval = atoi(scratch);
    }
    else
    {
      lst[i].interval = 1000;
    }
    if (nam_col >= 0) strncpy(lst[i].name,GetCsvColumnString(nam_col,str),32);
    if (fmt_col >= 0) strncpy(lst[i].dataformat,GetCsvColumnString(fmt_col,str),32);
    if (ipt_col >= 0) strncpy(lst[i].datainput,GetCsvColumnString(ipt_col,str),132);
    if (lim_col >= 0) strncpy(lst[i].datalimit,GetCsvColumnString(lim_col,str),64);
    if (rul_col >= 0) strncpy(lst[i].clbrrule_rcv,GetCsvColumnString(rul_col,str),132);
    if (rrcv_col >= 0) strncpy(lst[i].clbrrule_rcv,GetCsvColumnString(rrcv_col,str),132);
    if (rsnd_col >= 0) strncpy(lst[i].clbrrule_snd,GetCsvColumnString(rsnd_col,str),132);
    if (lng_col >= 0) strncpy(lst[i].map,GetCsvColumnString(lng_col,str),64);
    if (map_col >= 0) strncpy(lst[i].map,GetCsvColumnString(map_col,str),64);
    if (dsc_col >= 0) strncpy(lst[i].description,GetCsvColumnString(dsc_col,str),64);
    i++;
  }
  cdiDbEntryLst = lst;
  sizCdiEntryLst = i;
  capCdiEntryLst = n;
err:
  if (fp != NULL) fclose(fp);
  if (cc != 0)
  {
    if (lst != NULL) SystemFree(lst);
    sizCdiEntryLst = capCdiEntryLst = 0;
  }
  return cc;
}
int dbUpdate(int pid,int type,void *db,int siz)
{
   getDbRbList();
   if (siz == 0 && type != DB_ROLLBACK) type = DB_DUMP;
   switch (pid)
   {
     case PRP_MFDB:
       switch (type)
       {
         case DB_REPLACE: return useDbItemsAsMfDb((CdiMfDbEntry *)db,siz);
         case DB_UPDATE: return mergeMfDbItems((CdiMfDbEntry *)db,siz);
         case DB_ROLLBACK: return rollbackMfItems();
         case DB_DUMP:
         default: break;
       }
       return dumpMfDbItemList(NULL);
     case PRP_ADDRDB:
       switch (type)
       {
         case DB_REPLACE: return useDbItemsAsAddrDb((CdiDbEntry *)db,siz);
         case DB_UPDATE: return mergeDbItems((CdiDbEntry *)db,siz);
         case DB_ROLLBACK: return rollbackAddrItems();
         case DB_DUMP:
         default: break;
       }
       return dumpDbItemList(NULL);
     case PRP_SCHEDDB:
       switch (type)
       {
         case DB_REPLACE: return useDbItemsAsSchDb((CdiSchDbEntry *)db,siz);
         case DB_UPDATE: return mergeSchDbItems((CdiSchDbEntry *)db,siz);
         case DB_ROLLBACK: return rollbackSchItems();
         case DB_DUMP:
         default: break;
       }
       return dumpSchDbItemList(NULL);
     case PRP_SRVDB:
       switch (type)
       {
         case DB_REPLACE: return useDbItemsAsSrvDb((CdiSrvDbEntry *)db,siz);
         case DB_UPDATE: return mergeSrvDbItems((CdiSrvDbEntry *)db,siz);
         case DB_ROLLBACK: return rollbackSrvItems();
         case DB_DUMP:
         default: break;
       }
       return dumpSrvDbItemList(NULL);
     default:
       break;
   }
   return not_implemented;
}
int dbGet(int pid,void *db,UINT32 *siz)
{
   int dbsiz;
   if (siz == NULL) return invalid_parameter;
   if (*siz == 0) return dimension_error;
   dbsiz = (int)(*siz);
   switch (pid)
   {
     case PRP_MFDB:
       if (dbsiz > sizCdiMfEntryLst) dbsiz = sizCdiMfEntryLst;
       if (dbsiz > 0) memcpy(db,cdiMfDbEntryLst,dbsiz*sizeof(CdiMfDbEntry));
       *siz = (UINT32)dbsiz;
       return 0;
     case PRP_ADDRDB:
       if (dbsiz > sizCdiEntryLst) dbsiz = sizCdiEntryLst;
       if (dbsiz > 0) memcpy(db,cdiDbEntryLst,dbsiz*sizeof(CdiDbEntry));
       *siz = (UINT32)dbsiz;
       return 0;
     case PRP_SCHEDDB:
       if (dbsiz > sizCdiSchEntryLst) dbsiz = sizCdiSchEntryLst;
       if (dbsiz > 0) memcpy(db,cdiSchDbEntryLst,dbsiz*sizeof(CdiSchDbEntry));
       *siz = (UINT32)dbsiz;
       return 0;
     case PRP_SRVDB:
       if (dbsiz > sizCdiSrvEntryLst) dbsiz = sizCdiSrvEntryLst;
       if (dbsiz > 0) memcpy(db,cdiSrvDbEntryLst,dbsiz*sizeof(CdiSrvDbEntry));
       *siz = (UINT32)dbsiz;
       return 0;
     default:
       break;
   }
   return not_implemented;
}

CdiDbRb cdiDbRb;
CSVCOLUMN csvCdiDbRbColumns[] =
{
  { "ADDRESS","",0,CF_TEXT,128,(BYTE *)cdiDbRb.addr,NULL,0 },
  { "MANIFEST","",0,CF_TEXT,128,(BYTE *)cdiDbRb.mf,NULL,0 },
  { "SCHEDULE","",0,CF_TEXT,128,(BYTE *)cdiDbRb.sch,NULL,0 },
  { "SERVER","",0,CF_TEXT,128,(BYTE *)cdiDbRb.srv,NULL,0 },
};
CSVDB csvCdiRbDb =
{
  csvCdiDbRbColumns,
  sizeof(csvCdiDbRbColumns)/sizeof(csvCdiDbRbColumns[0]),
  0,
  NULL,
  NULL,
  NULL,
  0,
  NULL,
  0,
  0
};
int getDbRbList(void)
{
  int cc = csvReadFile(CdiDBpath,CDIRB_FILE,&csvCdiRbDb,NULL);
  if (cc != 0) memset(&cdiDbRb,0,sizeof(cdiDbRb));
  return cc;
}
int putDbRbList(void)
{
  if (csvCdiRbDb.siz == 0) csvCdiRbDb.siz = 1;
  return csvWriteFile(CdiDBpath,CDIRB_FILE,&csvCdiRbDb);
}
