#ifndef CDIDB_H_
#define CDIDB_H_

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#include "cdi.def"
typedef struct CdiDbEntry
{
  char bus[132]; /**< bus name  */
  char name[32]; /**< cdi device name */
  char description[96]; /**< device description  */
  char base[96]; /**< hdw base address ('Address' is composed of 'base' + 'params') */
  char params[96]; /**< hdw address parameters ('Address' is composed of 'base' + 'params') */
  char map[64]; /**< hdw address map ('Long Name' is synonym) */
  char access[64]; /**< bus access  */
  int line; /**< cdi line  */
  int number; /**< cdi device number ('0' for auto number) */
  int interval; /**< readback interval (default = '1000' ms) */
  char mask[64]; /**< data mask  */
  char pattern[64]; /**< target pattern (if applicable)  */
  char tolerance[16]; /**< scheduling tolerance  */
  char dataformat[32]; /**< data format (default = INT16)  */
  char datainput[132]; /**< data input  */
  char datalimit[64]; /**< data array limits (for compound read/write)  */
  char clbrrule_rcv[132]; /**< calibration rule for readback data  */
  char clbrrule_snd[132]; /**< calibration rule for input data  */
} CdiDbEntry;
extern CdiDbEntry *cdiDbEntryLst;
extern int sizCdiEntryLst;
extern int capCdiEntryLst;

typedef struct CdiMfDbEntry
{
  char bus[32];
  char lib[64];
  char env[128];
  int id;
  int simulation;
  int almThreshold;
  int reserved;
} CdiMfDbEntry;
extern CdiMfDbEntry *cdiMfDbEntryLst;
extern int sizCdiMfEntryLst;

typedef struct CdiSchDbEntry
{
  char prp[64];
  char dev[64];
  char tol[64];
} CdiSchDbEntry;
extern CdiSchDbEntry *cdiSchDbEntryLst;
extern int sizCdiSchEntryLst;

typedef struct CdiSrvDbEntry
{
  char prp[PROPERTY_NAME_SIZE];  /**< target extended property */
  char dev[DEVICE_NAME_SIZE];  /**< target extended device */
  char fmt[16];  /**< format string */
  char tol[64];  /**< schedule tolerance: empty(default) => no scheduling */
  int  siz; /**< target size: empty => registered size */
  int  tmr; /**< interval timer: empty => 1000 msec */
  int  hst; /**< TRUE: flag for a stock history else only use history.csv settings */
  int reserved;
} CdiSrvDbEntry;
extern CdiSrvDbEntry *cdiSrvDbEntryLst;
extern int sizCdiSrvEntryLst;

typedef struct DbMergeItem
{
  void *item; /**< reference to item */
  int idx; /**< start index in existing db */
  struct DbMergeItem *prv; /**< prev ptr */
  struct DbMergeItem *nxt; /**< next ptr */
} DbMergeItem;

typedef struct CdiDbRb
{
  char addr[128];
  char mf[128];
  char sch[128];
  char srv[128];
} CdiDbRb;
extern CdiDbRb cdiDbRb;

#define DB_REPLACE 0
#define DB_UPDATE 1
#define DB_ROLLBACK 2
#define DB_DUMP 3

#define CDIRB_FILE "cdirollback.csv"
#define CDIDBLOGFILE "cdidb.log"
#define DBITEM_TAG "CdiDbItem"
#define MFITEM_TAG "CdiMfItem"
#define SCHDITEM_TAG "CdiSchItem"
#define SRVITEM_TAG "CdiSrvItem"
#define DB_NICE_QUERY_SIZE 1000
#define MF_NICE_QUERY_SIZE 25
#define SCH_NICE_QUERY_SIZE 100
#define SRV_NICE_QUERY_SIZE 100
#define DB_OVERFLOW 100
#define CdiDbHdr "Name, Bus, Description, Address Base, Address Parameters, Address Map, Access, Line, Number, Interval, Mask, Pattern, Tolerance, Format, Input, Limit, Rule Recv, Rule Send"
#define CdiMfDbHdr "Bus, Library, Bus Env, Bus Id, Simulation, Alarm Threshold"
#define CdiSchDbHdr "Property, Device, Tolerance"
#define CdiSrvDbHdr "Property, Device, Format, Schedule Tolerance, Size, Interval, Keep History"

int populateDbItemList(void);
int populateMfItemList(void);
int populateSchItemList(void);
int populateSrvItemList(void);
int dumpDbItemList(DbMergeItem *mrg);
int dumpMfDbItemList(DbMergeItem *mrg);
int dumpSrvDbItemList(DbMergeItem *mrg);
int dumpSchDbItemList(DbMergeItem *mrg);
int registerDbStructs(void);
int mergeDbItems(CdiDbEntry *entries,int len);
int mergeMfDbItems(CdiMfDbEntry *entries,int len);
int useDbItemsAsAddrDb(CdiDbEntry *entries,int len);
int useDbItemsAsMfDb(CdiMfDbEntry *entries,int len);
int dbUpdate(int pid,int replace,void *db,int siz);
int dbGet(int pid,void *db,UINT32 *siz);
int getDbRbList(void);
int putDbRbList(void);
int addrdbcmp(CdiDbEntry *entries,int len);
int mfdbcmp(CdiMfDbEntry *entries,int len);
int srvdbcmp(CdiSrvDbEntry *entries,int len);
int schdbcmp(CdiSchDbEntry *entries,int len);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* CDI_H_ */
